searchData={"items":[{"type":"module","title":"InstructorLite","doc":"Main building blocks of InstructorLite.","ref":"InstructorLite.html"},{"type":"module","title":"Key Concepts - InstructorLite","doc":"Structured prompting can be quite different depending on the LLM and InstructorLite does only the bare minimum to abstract this complexity. This means the usage can be quite different depending on the adapter you're using, so make sure to consult adapter documentation to learn the details.\n\nThere are two key arguments used throughout this module. Understanding what they are will make your life a lot easier.\n\n* `params` - is an adapter-specific map, that contain values eventually sent to the LLM. More simply, this is the body that will be posted to the API endpoint. You prompt, model name, optional parameters like temperature all likely belong here.\n* `opts` - is a list of options that shape behavior of InstructorLite itself. Options may include things like which schema to cast response to, http client to use, api key, optional headers, http timeout, etc.","ref":"InstructorLite.html#module-key-concepts"},{"type":"module","title":"Shared options - InstructorLite","doc":"Most functions in this module accept a list of options.\n* `:response_model` - Required. A module implementing `InstructorLite.Instruction` behaviour, Ecto schema or [schemaless Ecto definition](https://hexdocs.pm/ecto/Ecto.Changeset.html#module-schemaless-changesets)\n\n* `:adapter` (`t:atom/0`) - A module implementing `InstructorLite.Adapter` behaviour. The default value is `InstructorLite.Adapters.OpenAI`.\n\n* `:max_retries` (`t:non_neg_integer/0`) - How many additional attempts to make if changeset validation fails. The default value is `0`.\n\n* `:validate_changeset` (function of arity 2) - Override function to be called instead of `response_model.validate_changeset/2` callback\n\n* `:notes` (`t:String.t/0`) - Additional notes about the schema that might be used by an adapter\n\n* `:json_schema` (`t:map/0`) - JSON schema to use instead of calling response_model.json_schema/0 callback or generating it at runtime using `InstructorLite.JSONSchema` module\n\n* `:adapter_context` (`t:term/0`) - Options used by adapter callbacks. See adapter docs for schema.\n\n* `:extra` (`t:term/0`) - Any arbitrary term for ad-hoc usage. For example, in `c:InstructorLite.Instruction.validate_changeset/2` callback","ref":"InstructorLite.html#module-shared-options"},{"type":"function","title":"InstructorLite.consume_response/3","doc":"Triage raw LLM response\n\nAttempts to cast raw response from `c:InstructorLite.Adapter.send_request/2` and either returns an object or an invalid changeset with new prompt that can be used for a retry.\n\nThis function will call `c:InstructorLite.Instruction.validate_changeset/2` callback, unless `validate_changeset` option is overridden in `opts`.","ref":"InstructorLite.html#consume_response/3"},{"type":"function","title":"InstructorLite.instruct/2","doc":"Perform instruction session from start to finish.\n\nThis function glues together all other functions and adds retries on top.","ref":"InstructorLite.html#instruct/2"},{"type":"function","title":"Examples - InstructorLite.instruct/2","doc":"","ref":"InstructorLite.html#instruct/2-examples"},{"type":"function","title":"Basic Example - InstructorLite.instruct/2","doc":"<!-- tabs-open -->","ref":"InstructorLite.html#instruct/2-basic-example"},{"type":"function","title":"OpenAI - InstructorLite.instruct/2","doc":"```\niex> InstructorLite.instruct(%{\n    messages: [\n      %{role: \"user\", content: \"John Doe is fourty two years old\"}\n    ]\n  },\n  response_model: %{name: :string, age: :integer},\n  adapter: InstructorLite.Adapters.OpenAI,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :openai_key)]\n)\n{:ok, %{name: \"John Doe\", age: 42}}\n```","ref":"InstructorLite.html#instruct/2-openai"},{"type":"function","title":"Anthropic - InstructorLite.instruct/2","doc":"```\niex> InstructorLite.instruct(%{\n    messages: [\n      %{role: \"user\", content: \"John Doe is fourty two years old\"}\n    ]\n  },\n  response_model: %{name: :string, age: :integer},\n  adapter: InstructorLite.Adapters.Anthropic,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :anthropic_key)]\n)\n{:ok, %{name: \"John Doe\", age: 42}}\n```","ref":"InstructorLite.html#instruct/2-anthropic"},{"type":"function","title":"Llamacpp - InstructorLite.instruct/2","doc":"```\niex> InstructorLite.instruct(%{\n    prompt: \"John Doe is fourty two years old\"\n  },\n  response_model: %{name: :string, age: :integer},\n  adapter: InstructorLite.Adapters.Llamacpp,\n  adapter_context: [url: Application.fetch_env!(:instructor_lite, :llamacpp_url)]\n)\n{:ok, %{name: \"John Doe\", age: 42}}\n```\n\n<!-- tabs-close -->","ref":"InstructorLite.html#instruct/2-llamacpp"},{"type":"function","title":"Using `max_retries` - InstructorLite.instruct/2","doc":"```\ndefmodule Rhymes do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n  \n  @primary_key false\n  embedded_schema do\n    field(:word, :string)\n    field(:rhymes, {:array, :string})\n  end\n  \n  @impl true\n  def validate_changeset(changeset, _opts) do\n    Ecto.Changeset.validate_length(changeset, :rhymes, is: 3)\n  end\nend\n\nInstructorLite.instruct(%{\n    messages: [\n      %{role: \"user\", content: \"Take the last word from the following line and add some rhymes to it\nEven though you broke my heart\"}\n    ]\n  },\n  response_model: Rhymes,\n  max_retries: 1,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :openai_key)]\n)\n{:ok, %Rhymes{word: \"heart\", rhymes: [\"part\", \"start\", \"dart\"]}}\n```","ref":"InstructorLite.html#instruct/2-using-max_retries"},{"type":"function","title":"InstructorLite.prepare_prompt/2","doc":"Prepare prompt that can be later sent to LLM\n\nThe prompt is added to `params`, so you need to cooperate with the adapter to know what you can provide there.\n\nThe function will call `c:InstructorLite.Instruction.notes/0` and `c:InstructorLite.Instruction.json_schema/0` callbacks for `response_model`. Both can be overriden with corresponding options in `opts`.","ref":"InstructorLite.html#prepare_prompt/2"},{"type":"type","title":"InstructorLite.opts/0","doc":"Options passed to instructor functions.","ref":"InstructorLite.html#t:opts/0"},{"type":"module","title":"InstructorLite.JSONSchema","doc":"Helper module to generate JSON Schema based on Ecto schema.\n\nJSON Schema comes in many flavors and different LLMs have different limitations. Currently, this module aims to implement a schema suitable for [OpenAI structured outputs](https://platform.openai.com/docs/guides/structured-outputs/supported-schemas). This type of JSON schema may not be optimal or compatible with other models. Therefore, it's recommended to use this module as a starting point to generate a schema during development. You can then manually refine the schema and either implement it in the `c:InstructorLite.Instruction.json_schema/0` callback or pass it as an option.\n\n> #### Warning {: .warning}\n>\n> For the reasons described above, neither backward compatibility nor compatibility with all built-in adapters are goals of this module.","ref":"InstructorLite.JSONSchema.html"},{"type":"function","title":"InstructorLite.JSONSchema.from_ecto_schema/1","doc":"Generates a JSON Schema from an Ecto schema.","ref":"InstructorLite.JSONSchema.html#from_ecto_schema/1"},{"type":"behaviour","title":"InstructorLite.Adapter","doc":"Behaviour for implementing adapter modules.\n\nThe role of an adapter module is to encapsulate all logic for a particular LLM, which can be quite specific. As a result, most of the details live in adapter modules and main InstructorLite interface is very broad.\n  \n> #### Implement your own! {: .tip}\n>\n> While built-in adapters are fairly flexible, users are encouraged to write their own adapters to establish a better control over prompt building, http clients, etc.","ref":"InstructorLite.Adapter.html"},{"type":"callback","title":"InstructorLite.Adapter.initial_prompt/2","doc":"Update `params` with model-specific prompt.","ref":"InstructorLite.Adapter.html#c:initial_prompt/2"},{"type":"callback","title":"InstructorLite.Adapter.parse_response/2","doc":"Parse API response.\n\nOn success, it returns an ok tuple with parsed json object ready to be cast to a changeset.\n\nOn error, it may return a 3-element error tuple with error reason. In the worst-case scenario, it returns the original input wrapped in an error tuple.","ref":"InstructorLite.Adapter.html#c:parse_response/2"},{"type":"callback","title":"InstructorLite.Adapter.retry_prompt/5","doc":"Update `params` with model-specific prompt for a retry. This usually involves appending errors from a changeset.","ref":"InstructorLite.Adapter.html#c:retry_prompt/5"},{"type":"callback","title":"InstructorLite.Adapter.send_request/2","doc":"Make request to model API.","ref":"InstructorLite.Adapter.html#c:send_request/2"},{"type":"type","title":"InstructorLite.Adapter.params/0","doc":"Map of adapter-specific values, such as messages, prompt, model name, or temperature. These parameters are typically sent to the LLM in request body.","ref":"InstructorLite.Adapter.html#t:params/0"},{"type":"type","title":"InstructorLite.Adapter.parsed_response/0","doc":"Parsed response content that can be cast to a changeset.","ref":"InstructorLite.Adapter.html#t:parsed_response/0"},{"type":"type","title":"InstructorLite.Adapter.response/0","doc":"Raw content of a successful response.","ref":"InstructorLite.Adapter.html#t:response/0"},{"type":"behaviour","title":"InstructorLite.Instruction","doc":"`use InstructorLite.Instruction` is a way to make your Ecto schema into an Instruction, which provides additional callbacks used by InstructorLite.","ref":"InstructorLite.Instruction.html"},{"type":"behaviour","title":"Example - InstructorLite.Instruction","doc":"```\ndefmodule SpamPrediction do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @notes \"\"\"\n  Field Descriptions:\n  - class: Whether or not the email is spam.\n  - reason: A short, less than 10-word rationalization for the classification.\n  - score: A confidence score between 0.0 and 1.0 for the classification.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl InstructorLite.Instruction\n  def validate_changeset(changeset, _opts) do\n    Ecto.Changeset.validate_number(changeset, :score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n```\n  \n> #### `use InstructorLite.Instruction` {: .info}\n>\n> When you `use InstructorLite.Instruction`, the Instruction module will set `@behaviour InstructorLite.Instruction`, and define default implementations of `c:notes/0` and `c:json_schema/0` callbacks.","ref":"InstructorLite.Instruction.html#module-example"},{"type":"callback","title":"InstructorLite.Instruction.json_schema/0","doc":"Defines JSON schema for the instruction.\n  \nBy default, `InstructorLite.JSONSchema.from_ecto_schema/1` is called at runtime every time InstructorLite needs to convert an Ecto schema to JSON schema. However, you can bake your own JSON schema into the `c:json_schema/0` callback to eliminate the need to do it on every call.\n\n> #### Tip {: .tip}\n> \n> Take advantage of this callback! Most JSON schemas are known ahead of time, so there is no need to constantly build them at runtime. In addition, `InstructorLite.JSONSchema` module aims to generate one-size-fits-all schema, so it's very unlikely to take full advantage of JSON capabilities of your LLM of choice.","ref":"InstructorLite.Instruction.html#c:json_schema/0"},{"type":"callback","title":"InstructorLite.Instruction.notes/0","doc":"Defines an optional free-form description of the schema.\n  \nYou can define a `@notes` attribute or `c:notes/0` callback to provide an additional free-form description to the schema. This description will be passed to the LLM together with the schema definition.","ref":"InstructorLite.Instruction.html#c:notes/0"},{"type":"callback","title":"InstructorLite.Instruction.validate_changeset/2","doc":"Called by `InstructorLite.consume_response/3` as part of changeset validation.\n\nIt has full access to all `opts`. If you need to pass an arbitrary term to this callback, use the `extra` key.","ref":"InstructorLite.Instruction.html#c:validate_changeset/2"},{"type":"callback","title":"Example - InstructorLite.Instruction.validate_changeset/2","doc":"```\n# Let's play a guessing game!\ndefmodule CoinGuess do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @primary_key false\n  embedded_schema do\n    field(:guess, Ecto.Enum, values: [:heads, :tails])\n  end\n\n  @impl InstructorLite.Instruction\n  def validate_changeset(cs, opts) do\n    target = Keyword.fetch!(opts, :extra)\n\n    case Ecto.Changeset.fetch_field!(cs, :guess) do\n      ^target -> cs\n      _ -> Ecto.Changeset.add_error(cs, :guess, \"Wrong! Try again\")\n    end\n  end\nend\n```","ref":"InstructorLite.Instruction.html#c:validate_changeset/2-example"},{"type":"module","title":"InstructorLite.Adapters.Anthropic","doc":"[Anthropic](https://docs.anthropic.com/en/home) adapter.\n\nThis adapter is implemented using the [Messages API](https://docs.anthropic.com/en/api/messages) and [function calling](https://docs.anthropic.com/en/docs/build-with-claude/tool-use).","ref":"InstructorLite.Adapters.Anthropic.html"},{"type":"module","title":"Params - InstructorLite.Adapters.Anthropic","doc":"`params` argument should be shaped as a [Create message request body](https://docs.anthropic.com/en/api/messages).","ref":"InstructorLite.Adapters.Anthropic.html#module-params"},{"type":"module","title":"Example - InstructorLite.Adapters.Anthropic","doc":"```\nInstructorLite.instruct(%{\n    messages: [%{role: \"user\", content: \"John is 25yo\"}],\n    model: \"claude-3-5-sonnet-20240620\",\n    metadata: %{user_id: \"3125\"}\n  },\n  response_model: %{name: :string, age: :integer},\n  adapter: InstructorLite.Adapters.Anthropic,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :anthropic_key)]\n)\n{:ok, %{name: \"John\", age: 25}}\n```","ref":"InstructorLite.Adapters.Anthropic.html#module-example"},{"type":"function","title":"InstructorLite.Adapters.Anthropic.initial_prompt/2","doc":"Updates `params` with prompt based on `json_schema` and `notes`.\n\nAlso specifies default `claude-3-5-sonnet-20240620` model and 1024 `max_tokens` if not provided by a user.","ref":"InstructorLite.Adapters.Anthropic.html#initial_prompt/2"},{"type":"function","title":"InstructorLite.Adapters.Anthropic.parse_response/2","doc":"Parse API response.\n\nCan return:\n  * `{:ok, parsed_object}` on success.\n  * `{:error, :unexpected_response, response}` if response is of unexpected shape.","ref":"InstructorLite.Adapters.Anthropic.html#parse_response/2"},{"type":"function","title":"InstructorLite.Adapters.Anthropic.retry_prompt/5","doc":"Updates `params` with prompt for retrying a request.\n\nThe error is represented as an erroneous `tool_result`.","ref":"InstructorLite.Adapters.Anthropic.html#retry_prompt/5"},{"type":"function","title":"InstructorLite.Adapters.Anthropic.send_request/2","doc":"Make request to Anthropic API","ref":"InstructorLite.Adapters.Anthropic.html#send_request/2"},{"type":"function","title":"Options - InstructorLite.Adapters.Anthropic.send_request/2","doc":"* `:api_key` (`t:String.t/0`) - Required. Anthropic API key\n\n* `:http_client` (`t:atom/0`) - Any module that follows `Req.post/2` interface The default value is `Req`.\n\n* `:http_options` (`t:keyword/0`) - Options passed to `http_client.post/2` The default value is `[receive_timeout: 60000]`.\n\n* `:url` (`t:String.t/0`) - API endpoint to use for sending requests The default value is `\"https://api.anthropic.com/v1/messages\"`.\n\n* `:version` (`t:String.t/0`) - Anthropic [API version](https://docs.anthropic.com/en/api/versioning) The default value is `\"2023-06-01\"`.","ref":"InstructorLite.Adapters.Anthropic.html#send_request/2-options"},{"type":"module","title":"InstructorLite.Adapters.Llamacpp","doc":"[LLaMA.cpp HTTP Server](https://github.com/ggerganov/llama.cpp/tree/master/examples/server) adapter.\n\nThis adapter is implemented using llama-specific [completion](https://github.com/ggerganov/llama.cpp/tree/master/examples/server#post-completion-given-a-prompt-it-returns-the-predicted-completion) endpoint.","ref":"InstructorLite.Adapters.Llamacpp.html"},{"type":"module","title":"Params - InstructorLite.Adapters.Llamacpp","doc":"`params` argument should be shaped as a [completion request body](https://github.com/ggerganov/llama.cpp/tree/master/examples/server#post-completion-given-a-prompt-it-returns-the-predicted-completion)","ref":"InstructorLite.Adapters.Llamacpp.html#module-params"},{"type":"module","title":"Example - InstructorLite.Adapters.Llamacpp","doc":"```\nInstructorLite.instruct(%{\n    prompt: \"John is 25yo\",\n    temperature: 0.8\n  },\n  response_model: %{name: :string, age: :integer},\n  adapter: InstructorLite.Adapters.Llamacpp,\n  adapter_context: [url: \"http://localhost:8000/completion\"]\n)\n{:ok, %{name: \"John\", age: 25}}\n```","ref":"InstructorLite.Adapters.Llamacpp.html#module-example"},{"type":"function","title":"InstructorLite.Adapters.Llamacpp.initial_prompt/2","doc":"Updates `params` with prompt based on `json_schema` and `notes`.\n\nIt uses `json_schema` and `system_prompt` request parameters.","ref":"InstructorLite.Adapters.Llamacpp.html#initial_prompt/2"},{"type":"function","title":"InstructorLite.Adapters.Llamacpp.parse_response/2","doc":"Parse API response.\n\nCan return:\n  * `{:ok, parsed_json}` on success.\n  * `{:error, :unexpected_response, response}` if response is of unexpected shape.","ref":"InstructorLite.Adapters.Llamacpp.html#parse_response/2"},{"type":"function","title":"InstructorLite.Adapters.Llamacpp.retry_prompt/5","doc":"Updates `params` with prompt for retrying a request.","ref":"InstructorLite.Adapters.Llamacpp.html#retry_prompt/5"},{"type":"function","title":"InstructorLite.Adapters.Llamacpp.send_request/2","doc":"Make request to llamacpp HTTP server.","ref":"InstructorLite.Adapters.Llamacpp.html#send_request/2"},{"type":"function","title":"Options - InstructorLite.Adapters.Llamacpp.send_request/2","doc":"* `:http_client` (`t:atom/0`) - Any module that follows `Req.post/2` interface The default value is `Req`.\n\n* `:http_options` (`t:keyword/0`) - Options passed to `http_client.post/2` The default value is `[receive_timeout: 60000]`.\n\n* `:url` (`t:String.t/0`) - Required. API endpoint to use, for example `http://localhost:8000/completion`","ref":"InstructorLite.Adapters.Llamacpp.html#send_request/2-options"},{"type":"module","title":"InstructorLite.Adapters.OpenAI","doc":"[OpenAI](https://platform.openai.com/docs/overview) adapter.\n\nThis adapter is implemented using chat completions endpoint and [structured outputs](https://platform.openai.com/docs/guides/structured-outputs/structured-outputs).","ref":"InstructorLite.Adapters.OpenAI.html"},{"type":"module","title":"Params - InstructorLite.Adapters.OpenAI","doc":"`params` argument should be shaped as a [Create chat completion request body](https://platform.openai.com/docs/api-reference/chat/create).","ref":"InstructorLite.Adapters.OpenAI.html#module-params"},{"type":"module","title":"Example - InstructorLite.Adapters.OpenAI","doc":"```\nInstructorLite.instruct(%{\n    messages: [%{role: \"user\", content: \"John is 25yo\"}],\n    model: \"gpt-4o-mini\",\n    service_tier: \"default\"\n  },\n  response_model: %{name: :string, age: :integer},\n  adapter: InstructorLite.Adapters.OpenAI,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :openai_key)]\n)\n{:ok, %{name: \"John\", age: 25}}\n```","ref":"InstructorLite.Adapters.OpenAI.html#module-example"},{"type":"function","title":"InstructorLite.Adapters.OpenAI.initial_prompt/2","doc":"Updates `params` with prompt based on `json_schema` and `notes`.\n\nAlso specifies default `gpt-4o-mini` model if not provided by a user.","ref":"InstructorLite.Adapters.OpenAI.html#initial_prompt/2"},{"type":"function","title":"InstructorLite.Adapters.OpenAI.parse_response/2","doc":"Parse chat completion endpoint response.\n\nCan return:\n  * `{:ok, parsed_json}` on success.\n  * `{:error, :refusal, reason}` on [refusal](https://platform.openai.com/docs/guides/structured-outputs/refusals).\n  * `{:error, :unexpected_response, response}` if response is of unexpected shape.","ref":"InstructorLite.Adapters.OpenAI.html#parse_response/2"},{"type":"function","title":"InstructorLite.Adapters.OpenAI.retry_prompt/5","doc":"Updates `params` with prompt for retrying a request.","ref":"InstructorLite.Adapters.OpenAI.html#retry_prompt/5"},{"type":"function","title":"InstructorLite.Adapters.OpenAI.send_request/2","doc":"Make request to OpenAI API.","ref":"InstructorLite.Adapters.OpenAI.html#send_request/2"},{"type":"function","title":"Options - InstructorLite.Adapters.OpenAI.send_request/2","doc":"* `:api_key` (`t:String.t/0`) - Required. OpenAI API key\n\n* `:http_client` (`t:atom/0`) - Any module that follows `Req.post/2` interface The default value is `Req`.\n\n* `:http_options` (`t:keyword/0`) - Options passed to `http_client.post/2` The default value is `[receive_timeout: 60000]`.\n\n* `:url` (`t:String.t/0`) - API endpoint to use for sending requests The default value is `\"https://api.openai.com/v1/chat/completions\"`.","ref":"InstructorLite.Adapters.OpenAI.html#send_request/2-options"},{"type":"extras","title":"InstructorLite","doc":"# InstructorLite\n\n[![Instructor version](https://img.shields.io/hexpm/v/instructor_lite.svg)](https://hex.pm/packages/instructor_lite)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-lightgreen.svg)](https://hexdocs.pm/instructor_lite/)\n[![Hex Downloads](https://img.shields.io/hexpm/dt/instructor_lite)](https://hex.pm/packages/instructor_lite)\n[![GitHub stars](https://img.shields.io/github/stars/martosaur/instructor_lite.svg)](https://github.com/martosaur/instructor_lite/stargazers)\n[![CI](https://github.com/martosaur/instructor_lite/actions/workflows/ci.yml/badge.svg)](https://github.com/martosaur/instructor_lite/actions/workflows/ci.yml)\n\nStructured prompting for LLMs. InstructorLite is a fork and spiritual successor to [instructor_ex](https://github.com/thmsmlr/instructor_ex) library, which is the Elixir member of the great [Instructor](https://useinstructor.com/) family.\n \nThe Instructor is useful for coaxing an LLM to return JSON that maps to an Ecto schema that you provide, rather than the default unstructured text output. If you define your own validation logic, Instructor can automatically retry prompts when validation fails (returning natural language error messages to the LLM, to guide it when making corrections).","ref":"readme.html"},{"type":"extras","title":"Why Lite - InstructorLite","doc":"InstructorLite is designed to be:\n1. **Lean**. It does so little it makes you question if you should just write your own version!\n2. **Composable**. Almost everything it does can be overridden or extended.\n3. **Magic-free**. It doesn't hide complexity behind one line function calls, but does its best to provide you with enough information to understand what's going on.\n\nInstructorLite comes with 3 adapters: [OpenAI](https://openai.com/api/), [Anthropic](https://www.anthropic.com/) and [Llamacpp](https://github.com/ggerganov/llama.cpp).","ref":"readme.html#why-lite"},{"type":"extras","title":"Features - InstructorLite","doc":"InstructorLite can be boiled down to these features:\n1. It provides a very simple function for generating JSON-schema from Ecto schema.\n2. It facilitates generating prompts, calling LLMs, casting and validating responses, including retrying prompts when validation fails.\n3. It holds knowledge of major LLM providers' API interfaces with adapters.\n\nAny of the features above can be used independently.","ref":"readme.html#features"},{"type":"extras","title":"Usage - InstructorLite","doc":"Define an instruction, which is a normal Ecto schema with an extra `use Instructor.Instruction` call.\n\n```elixir\ndefmodule UserInfo do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n  \n  @primary_key false\n  embedded_schema do\n    field(:name, :string)\n    field(:age, :integer)\n  end\nend\n```\n\nNow let's use `InstructorLite.instruct/2` to fill the schema from unstructured text:\n\n<!-- tabs-open -->","ref":"readme.html#usage"},{"type":"extras","title":"OpenAI - InstructorLite","doc":"```elixir\niex> InstructorLite.instruct(%{\n    messages: [\n      %{role: \"user\", content: \"John Doe is fourty two years old\"}\n    ]\n  },\n  response_model: UserInfo,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :openai_key)]\n)\n{:ok, %UserInfo{name: \"John Doe\", age: 42}}\n```","ref":"readme.html#openai"},{"type":"extras","title":"Anthropic - InstructorLite","doc":"```elixir\niex> InstructorLite.instruct(%{\n    messages: [\n      %{role: \"user\", content: \"John Doe is fourty two years old\"}\n    ]\n  },\n  response_model: UserInfo,\n  adapter: InstructorLite.Adapters.Anthropic,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :anthropic_key)]\n)\n{:ok, %UserInfo{name: \"John Doe\", age: 42}}\n```","ref":"readme.html#anthropic"},{"type":"extras","title":"Llamacpp - InstructorLite","doc":"```elixir\niex> InstructorLite.instruct(%{\n    prompt: \"John Doe is fourty two years old\"\n  },\n  response_model: UserInfo,\n  adapter: InstructorLite.Adapters.Llamacpp,\n  adapter_context: [url: Application.fetch_env!(:instructor_lite, :llamacpp_url)]\n)\n{:ok, %UserInfo{name: \"John Doe\", age: 42}}\n```\n\n<!-- tabs-close -->","ref":"readme.html#llamacpp"},{"type":"extras","title":"Configuration - InstructorLite","doc":"InstructorLite _does not_ access the application environment for configuration options like adapter or api key. Instead, they're passed as options when needed. Note that different adapters may require different options, so make sure to check their documentation.","ref":"readme.html#configuration"},{"type":"extras","title":"Installation - InstructorLite","doc":"In your mix.exs, add `:instructor_lite` to your list of dependencies:\n\n```elixir\ndef deps do\n  [\n    {:instructor_lite, \"~> 0.1.0\"}\n  ]\nend\n```\n\nOptionally, include `m:Req` HTTP client as it's used by default:\n\n```elixir\ndef deps do\n  [\n    {:req, \"~> 0.5 or ~> 1.0\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Philosophy & Motivations","doc":"# Philosophy & Motivations\n\nInstructor is simple library that allows you to maximally leverage large language models (LLMs).\nIt acts as your bridge between [Karpathy's Software 1.0 and Software 2.0](https://karpathy.medium.com/software-2-0-a64152b37c35).\n\nSoftware 1.0 is the world we're all used to. \nIt's datastructures, databases, algorithms, code.\n\nSoftware 2.0 on the other hand is neural networks.\nThey're models to problems for which it's easier to gather examples of desired behavior than describe how to produce it.","ref":"philosophy.html"},{"type":"extras","title":"A New Modality - Philosophy & Motivations","doc":"Software 2.0, as it exists today operates in the modality of text, images, and audio.\nAll modalities that have been traditionally very difficult to deal with in Software 1.0.\n\nWe now have these things called LLMs.\nFor the first time we can solve non-trivial `f(text) -> text` problems.\n\n```\n\n  llm(\"Recipe for a banana smoothie\") -> \"\"\"\n    A banana smoothie is a delicious and nutritious treat that's quite easy to make. Here's a simple recipe for you:\n\n    Ingredients:\n\n    2 ripe bananas\n    1 cup milk (you can use almond, soy, or cow's milk)\n    1/2 cup Greek yogurt (optional, for extra creaminess)\n    1-2 tablespoons honey or maple syrup (adjust according to taste)\n    A pinch of cinnamon (optional)\n    Ice cubes (optional, for a colder smoothie)\n    Instructions:\n\n    Peel the bananas and place them in a blender.\n    Add the milk, Greek yogurt (if using), honey/maple syrup, and cinnamon.\n    Add a few ice cubes if you prefer a colder smoothie.\n    Blend until smooth and creamy. If the smoothie is too thick, you can add a little more milk to reach your desired consistency.\n    Taste and adjust the sweetness if necessary.\n    Pour into glasses and serve immediately.\n\n    Enjoy your banana smoothie!\n  \"\"\"\n\n```\n\nHowever from a Software 1.0 perspective, this isn't terribly useful.\nRemember, Software 1.0 doesn't have robust, non trivial semantic text processing.\nWe can pretty easily create the function's arguments, but are you sure you can robustly parse the output? Unlikely.","ref":"philosophy.html#a-new-modality"},{"type":"extras","title":"LLMs aren't only for chatbots - Philosophy & Motivations","doc":"Over the past year you've probably noticed that every LLM company produces some variation of a chatbot -\nChat with GPT, Chat with your PDF, Chat with your AI Girlfriend.\n\nThis makes sense, when you think about it.\nWe're really good at creating strings in Software 1.0 from structured data, but we're terrible at creating structured data from strings.\nNor are we any good at interpreting the semantic meaning of strings.\n\nTherefore, it only makes sense to design experiences where we present them to the user for interpretation.\nAlas, our ubiquitous chatbot interface.\n\nBut it doesn't have to be that way. \nWhat if we can get our LLMs to respond with structured data?\nThat'd give us the all our domains and ranges covered,\n\n```\n  f(text) -> text\n  f(data_structure |> to_string) -> text\n  f(data_structure |> to_string) -> data_structure\n```\n\nWith such a capability, you would have full bidirectional interoperability between software 1.0 and software 2.0. \nThat's what Instructor provides.","ref":"philosophy.html#llms-aren-t-only-for-chatbots"},{"type":"extras","title":"It's just Ecto baby - Philosophy & Motivations","doc":"Now to do this we need to give the LLM a schema that it has to conform to in its output text. \nIf we can guarantee that it outputs JSON that matches this schema, we can treat it as any other user data we're used to in Elixir.\n\nThat's why, we figure, let's just build this on top of Ecto. It's already in your app after all.\nYou already know how to use it.\nIt already provides a robust validations API.\nLet's just treat the LLM as if it's one of your users, submitting data through a form.\n\n```elixir\ndefmodule Recipe do\n  use Ecto.Schema\n\n  @doc \"\"\"\n  Our AI generated delicious recipe.\n  \"\"\"\n  @primary_key false\n  embedded_schema do \n    field :title, :string\n    field :cook_time, :integer\n    field :steps, {:array, :string}\n    embeds_many :ingredients, Ingredients, primary_key: false do\n      field :name, :string\n      field :quantity, :decimal\n      field :unit, :string\n    end\n  end\nend\n```\n\nFrom this instructor takes your ecto schema, converts it into a JSON schema, and passes it on to the LLM.\nIt uses tools provided by LLMs like function calling and structured output to ensure that we get back a result that is exactly matching our Ecto schema.\nAt which point it's all just Ecto code.\nWe cast the fields, we do the validations, and we return the result back to you.\n\n```elixir\nInstructorLite.instruct(%{\n    messages: [\n      %{ role: \"user\", content: \"Give me a recipe for a banana smoothie\" }\n    ]\n  },\n  response_model: Recipe,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :openai_key)]\n)\n\n# => {:ok, %Recipe{title: \"Banana smoothie\", cook_time: 15, steps: [...], ingredients: [...]}\n```\n\nInstructor makes Ecto is our interchange format between Software 1.0 and 2.0.\nPerfect because Ecto is a mainstay in the ecosystem and compatible with so many other libraries.","ref":"philosophy.html#it-s-just-ecto-baby"},{"type":"extras","title":"Used Anywhere where Ecto is used - Philosophy & Motivations","doc":"Now this is a neat capability, but you may be thinking, where can I use this?  \nThis is Virgin Territory. We've never had this capability. It is a new modality in computing.\nIt will require a bit of imagination.\nA good heuristic to start with is, \"wherever I use Ecto, I can use Instructor.\" \n\n```\n  user -> instructor -> code\n  code -> instructor -> code\n  code -> instructor -> user\n```\n\nThere is a 40Gb LLM file sitting on Huggingface that encodes the entirety of human knowledge, you now have a query interface for it.","ref":"philosophy.html#used-anywhere-where-ecto-is-used"},{"type":"extras","title":"Migrating from Instructor","doc":"# Migrating from Instructor\n\n`InstructorLite` is not meant as a drop-in replacement for `Instructor`, so there are a number of things to consider when making a decision to switch.\n\n✅ You might want to switch if:\n* You need more customization (e.g. using `Tesla` instead of `Req`)\n* You want to have fine-grained control over side-effects (e.g. delay executing requests to use [Batch API](https://platform.openai.com/docs/guides/batch))\n* You plan to tinker with the library's source code or write your own adapters\n\n❌ You might not want to switch if:\n* Instructor already works for you and you want things to continue to \"just work\"\n* You need streaming capabilities","ref":"migrating_from_instructor.html"},{"type":"extras","title":"Migrating Schemas - Migrating from Instructor","doc":"Instructor works with any Ecto schema and optionally allows you to use `Instructor.Validator` behaviour for the `Instructor.Validator.validate_changeset/2` callback.\n\nInstructorLite can also work with any Ecto schema, but it doesn't use the `@doc` attribute for semantic description. Instead, you can `use InstructorLite.Instruction` and define the `@notes` attribute. The `c:InstructorLite.Instruction.validate_changeset/2` callback is still here, but it's always of 2 arity.\n\nAs a result, you don't have to add a docstring in your release.\n\n<!-- tabs-open -->","ref":"migrating_from_instructor.html#migrating-schemas"},{"type":"extras","title":"Instructor - Migrating from Instructor","doc":"```elixir\ndefmodule SpamPrediction do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @doc \"\"\"","ref":"migrating_from_instructor.html#instructor"},{"type":"extras","title":"Field Descriptions: - Migrating from Instructor","doc":"- class: Whether or not the email is spam.\n  - reason: A short, less than 10 word rationalization for the classification.\n  - score: A confidence score between 0.0 and 1.0 for the classification.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> Ecto.Changeset.validate_number(:score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n```\n\n```elixir\n# mix.exs\ndef project do\n  # ...\n  releases: [\n    myapp: [\n      strip_beams: [keep: [\"Docs\"]]\n    ]\n  ]\nend\n```","ref":"migrating_from_instructor.html#field-descriptions"},{"type":"extras","title":"InstructorLite - Migrating from Instructor","doc":"```elixir\ndefmodule SpamPrediction do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @notes \"\"\"","ref":"migrating_from_instructor.html#instructorlite"},{"type":"extras","title":"Field Descriptions: - Migrating from Instructor","doc":"- class: Whether or not the email is spam.\n  - reason: A short, less than 10 word rationalization for the classification.\n  - score: A confidence score between 0.0 and 1.0 for the classification.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl InstructorLite.Instruction\n  def validate_changeset(changeset, _opts) do\n    changeset\n    |> Ecto.Changeset.validate_number(:score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n```\n\n<!-- tabs-close -->","ref":"migrating_from_instructor.html#field-descriptions"},{"type":"extras","title":"Migrating `Instructor.chat_completions/2` - Migrating from Instructor","doc":"`Instructor.chat_completion/2` works primarily with `params`, but also accepts an optional `config` as the second argument. Some parameters are passed to the LLM and some control Instructor's own behaviour.\n\nThe InstructorLite equivalent function is `InstructorLite.instruct/2`. It also accepts `params` as the first argument, but many params were moved to the second `opts` argument. Read more about this in [Key Concepts](`m:InstructorLite#key-concepts`). \n\nIn addition, InstructorLite does not access application configuration, so API keys are typically passed through `opts`.\n\n\n<!-- tabs-open -->","ref":"migrating_from_instructor.html#migrating-instructor-chat_completions-2"},{"type":"extras","title":"Instructor - Migrating from Instructor","doc":"```elixir\nInstructor.chat_completion(%{\n  model: \"gpt-3.5-turbo\",\n  response_model: SpamPrediction,\n  messages: [\n    %{\n      role: \"user\",\n      content: \"Classify the following text: Hello, I am a Nigerian prince and I would like to give you $1,000,000.\"\n    }\n  ]\n})\n{:ok, %SpamPrediction{class: :spam, score: 0.999}}\n```\n\n```elixir\n# config.exs\nconfig :instructor, \n  adapter: Instructor.Adapters.OpenAI,\n  openai: [api_key: \"my_secret_key\"]\n```","ref":"migrating_from_instructor.html#instructor"},{"type":"extras","title":"InstructorLite - Migrating from Instructor","doc":"```elixir\nInstructorLite.instruct(%{\n    messages: [\n      %{\n        role: \"user\",\n        content: \"Classify the following text: Hello, I am a Nigerian prince and I would like to give you $1,000,000.\"\n      }\n    ],\n    model: \"gpt-3.5-turbo\"\n  },\n  response_model: Instructor.Demos.SpamPrediction,\n  adapter: InstructorLite.Adapters.OpenAI,\n  adapter_context: [api_key: \"my_secret_key\"]\n})\n{:ok, %SpamPrediction{class: :spam, score: 0.999}}\n```\n\n<!-- tabs-close -->","ref":"migrating_from_instructor.html#instructorlite"},{"type":"extras","title":"Migrating Custom Ecto Types - Migrating from Instructor","doc":"`Instructor.EctoType` allows you to define a way to dump a custom Ecto type to JSON schema using the `Instructor.EctoType.to_json_schema/0` callback.\n\nInstructorLite lowers the bar for the built-in Ecto-to-JSON-schema converter. It only works for basic Ecto types and may not be compatible with all LLMs. As soon as basic JSON schema is not enough for you, you should define the schema manually using the `c:InstructorLite.Instruction.json_schema/0` callback.\n\n\n<!-- tabs-open -->","ref":"migrating_from_instructor.html#migrating-custom-ecto-types"},{"type":"extras","title":"Instructor - Migrating from Instructor","doc":"```elixir\ndefmodule Ecto.CSVDataFrame do\n  use Ecto.Type\n  use Instructor.EctoType\n\n  def type, do: :string\n\n  def cast(csv_str) when is_binary(csv_str) do\n    df = Explorer.DataFrame.load_csv!(csv_str)\n    {:ok, df}\n  end\n\n  def cast(%Explorer.DataFrame{} = df), do: {:ok, df}\n  def cast(_), do: :error\n\n  def to_json_schema(),\n    do: %{\n      type: \"string\",\n      description: \"A CSV representation of a data table\"\n    }\n\n  def dump(x), do: {:ok, x}\n  def load(x), do: {:ok, x}\nend\n```\n\n```elixir\ndefmodule Database do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @doc \"\"\"\n  The extracted database will contain one or more tables with data as csv formatted with ',' delimiters\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    embeds_many :tables, DataFrame, primary_key: false do\n      field(:name, :string)\n      field(:data, Ecto.CSVDataFrame)\n    end\n  end\nend\n```","ref":"migrating_from_instructor.html#instructor"},{"type":"extras","title":"InstructorLite - Migrating from Instructor","doc":"```elixir\ndefmodule Ecto.CSVDataFrame do\n  use Ecto.Type\n\n  def type, do: :string\n\n  def cast(csv_str) when is_binary(csv_str) do\n    df = Explorer.DataFrame.load_csv!(csv_str)\n    {:ok, df}\n  end\n\n  def cast(%Explorer.DataFrame{} = df), do: {:ok, df}\n  def cast(_), do: :error\n\n  def dump(x), do: {:ok, x}\n  def load(x), do: {:ok, x}\nend\n```\n\n```elixir\ndefmodule Database do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @notes \"\"\"\n  The extracted database will contain one or more tables with data as csv formatted with ',' delimiters\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    embeds_many :tables, DataFrame, primary_key: false do\n      field(:name, :string)\n      field(:data, Ecto.CSVDataFrame)\n    end\n  end\n  \n  @impl InstructorLite.Instruction\n  def json_schema do\n    %{\n      type: \"object\",\n      description: \"\",\n      title: \"Database\",\n      required: [:tables],\n      \"$defs\": %{\n        \"DataFrame\" => %{\n          type: \"object\",\n          description: \"\",\n          title: \"DataFrame\",\n          required: [:data, :name],\n          properties: %{data: %{type: \"string\"}, name: %{type: \"string\"}},\n          additionalProperties: false\n        }\n      },\n      properties: %{tables: %{type: \"array\", items: %{\"$ref\": \"#/$defs/DataFrame\"}}},\n      additionalProperties: false\n    }\n  end\nend\n```\n\n<!-- tabs-close -->","ref":"migrating_from_instructor.html#instructorlite"},{"type":"extras","title":"Migrating `Instructor.validate_with_llm/2` - Migrating from Instructor","doc":"`Instructor.Validator.validate_with_llm/2` is a nifty function that allows you to use LLM to validate your changeset.\n\nInstructorLite doesn't have this function. Instead, you can write your own thin helper function.\n\n<!-- tabs-open -->","ref":"migrating_from_instructor.html#migrating-instructor-validate_with_llm-2"},{"type":"extras","title":"Instructor - Migrating from Instructor","doc":"```elixir\ndefmodule QuestionAnswer do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @primary_key false\n  embedded_schema do\n    field(:question, :string)\n    field(:answer, :string)\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> validate_with_llm(:answer, \"Do not say anything objectionable\")\n  end\nend\n```\n\n```elixir\n%QuestionAnswer{}\n|> Instructor.cast_all(%{\n  question: \"What is the meaning of life?\",\n  answer: \"Sex, drugs, and rock'n roll\"\n})\n|> QuestionAnswer.validate_changeset()\n\n#Ecto.Changeset ,\n  valid?: false\n>\n```","ref":"migrating_from_instructor.html#instructor"},{"type":"extras","title":"InstructorLite - Migrating from Instructor","doc":"```elixir\ndefmodule QuestionAnswer do\n  use Ecto.Schema\n\n  @primary_key false\n  embedded_schema do\n    field :question, :string\n    field :answer, :string\n  end\n\n  def changeset(data, params) do\n    data\n    |> Ecto.Changeset.cast(params, [:answer])\n    |> validate_with_llm(:answer, \"do not say anything objectionable\")\n  end\n\n  def validate_with_llm(changeset, field, rule) do\n    Ecto.Changeset.validate_change(changeset, field, fn ^field, value ->\n      %{\n        messages: [\n          %{\n            role: \"system\",\n            content: \"\"\"\n            You are a world-class validation model. Capable of determining if the following value is valid for the statement, if it is not, explain why.\n            \"\"\"\n          },\n          %{\n            role: \"user\",\n            content: \"Does `#{value}` follow the rule: #{rule}\"\n          }\n        ]\n      }\n      |> InstructorLite.instruct(\n        response_model: %{valid?: :boolean, reason: :string},\n        adapter_context: [api_key: \"my_api_token\"]\n      )\n      |> case do\n        {:ok, %{valid?: false, reason: reason}} ->\n          [{field, reason}]\n\n        _ -> []\n      end\n    end)\n  end\nend\n```\n\n```elixir\n%QuestionAnswer{question: \"What is the meaning of life?\"}\n|> QuestionAnswer.changeset(%{answer: \"Sex, drugs, and rock'n roll\"})\n\n#Ecto.Changeset ,\n  valid?: false,\n  ...\n>\n```\n\n<!-- tabs-close -->","ref":"migrating_from_instructor.html#instructorlite"},{"type":"extras","title":"Migrating Llamacpp Usage - Migrating from Instructor","doc":"Instructor's Llamacpp was implemented through converting JSON schema to GBNF grammar.\n\nInstructorLite offers a very basic LLamacpp adapter which passes JSON schema to the Llamacpp server. If you want to use Llamacpp with InstructorLite, you will likely need to learn how your model of choice works and write your own adapter. InstructorLite can offer you a flexible enough chassis to build upon, but doesn't have much expertise to share 🫡.","ref":"migrating_from_instructor.html#migrating-llamacpp-usage"},{"type":"extras","title":"Changelog","doc":"# Changelog","ref":"changelog.html"},{"type":"extras","title":"v0.1.0 - Changelog","doc":"* Initial release","ref":"changelog.html#v0-1-0"},{"type":"extras","title":"Text Classification","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Text Classification\n\n```elixir\nMix.install(\n  [\n    {:instructor_lite, path: Path.expand(\"../../\", __DIR__)},\n    {:req, \"~> 0.5\"}\n  ]\n)\n```","ref":"text-classification.html"},{"type":"extras","title":"Motivation - Text Classification","doc":"Text classification is a common task in NLP and broadly applicable across software. Whether it be spam detection, or support ticket categorization, NLP is at the core. Historically, this required training custom, bespoke models that required collecting thousands of pre-labeled examples. With LLMs a lot of this knowledge is already encoded into the model. With proper instruction and guiding the output to a known set of classifications using GPT you can be up and running with a text classification model in no time.\n\nHell, you can even use instructor to help generate the training set to train your own more efficient model. But let's not get ahead of ourselves, there's more on that later in the tutorials.","ref":"text-classification.html#motivation"},{"type":"extras","title":"Setup - Text Classification","doc":"In order to run code in this notebook, you need to add your OpenAI API key as an `OPENAI_KEY` Livebook secret. It will then be accessible through an environmental variable.\n\n```elixir\nsecret_key = System.fetch_env!(\"LB_OPENAI_KEY\")\n:ok\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n:ok\n```","ref":"text-classification.html#setup"},{"type":"extras","title":"Binary Text Classification - Text Classification","doc":"Spam detection is a classic example of binary text classification. It's as simple as returning a true / false of whether an example is in the class. This is pretty trivial to implement in instructor.\n\n````elixir\ndefmodule SpamPrediction do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @notes \"\"\"","ref":"text-classification.html#binary-text-classification"},{"type":"extras","title":"Field Descriptions: - Text Classification","doc":"- class: Whether or not the email is spam.\n  - reason: A short, less than 10 word rationalization for the classification.\n  - score: A confidence score between 0.0 and 1.0 for the classification.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl true\n  def validate_changeset(changeset, _opts) do\n    changeset\n    |> Ecto.Changeset.validate_number(:score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n\nis_spam? = fn text ->\n  InstructorLite.instruct(%{\n      model: \"gpt-4o-mini\",\n      messages: [\n        %{\n          role: \"user\",\n          content: \"\"\"\n          Your purpose is to classify customer support emails as either spam or not.\n          This is for a clothing retail business.\n          They sell all types of clothing.\n  \n          Classify the following email: \n          ```\n          #{text}\n          ```\n          \"\"\"\n        }\n      ]\n    },\n    response_model: SpamPrediction,\n    max_retries: 1,\n    adapter_context: [api_key: secret_key]\n  )\nend\n\nis_spam?.(\"Hello I am a Nigerian prince and I would like to send you money\")\n````\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok, %SpamPrediction{class: :spam, reason: \"Common spam trope about money\", score: 0.95}}\n```\n\nWe don't have to stop just at a boolean inclusion, we can also easily extend this idea to multiple categories or classes that we can classify the text into. In this example, let's consider classifying support emails. We want to know whether it's a `general_inquiry`, `billing_issue`, or a `technical_issue` perhaps it rightly fits in multiple classes. This can be useful if we want to cc' specialized support agents when intersecting customer issues occur\n\nWe can leverage `Ecto.Enum` to define a schema that restricts the LLM output to be a list of those values. We can also provide a `@notes` description to help guide the LLM with the semantic understanding of what these classifications ought to represent.\n\n```elixir\ndefmodule EmailClassifications do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @notes \"\"\"\n  A classification of a customer support email.\n\n  technical_issue - whether the user is having trouble accessing their account.\n  billing_issue - whether the customer is having trouble managing their billing or credit card\n  general_inquiry - all other issues\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:tags, {:array, Ecto.Enum},\n      values: [:general_inquiry, :billing_issue, :technical_issue]\n    )\n  end\nend\n\nclassify_email = fn text ->\n  {:ok, %{tags: result}} =\n    InstructorLite.instruct(%{\n        messages: [\n          %{\n            role: \"user\",\n            content: \"Classify the following text: #{text}\"\n          }\n        ]\n      },\n      response_model: EmailClassifications,\n      adapter_context: [api_key: secret_key]\n    )\n\n  result\nend\n\nclassify_email.(\"My account is locked and I can't access my billing info.\")\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[:technical_issue, :billing_issue]\n```\n\n<!-- livebook:{\"offset\":4685,\"stamp\":{\"token\":\"XCP.piCFFvM4UdfB4LjkjwMf2_8-2xjr87WuqUlbcBvMMI-bUvGSQ-Tx5Ulp88x9jlZhxy6TIigBop8tk2OVP3E_vm7nFV-QYXxQHAO5ydrbjVczreUnxQ\",\"version\":2}} -->","ref":"text-classification.html#field-descriptions"},{"type":"extras","title":"Computer Vision - Extracting Data from Images","doc":"<!-- livebook:{\"file_entries\":[{\"name\":\"shopify-screenshot.png\",\"type\":\"attachment\"}],\"persist_outputs\":true} -->\n\n# Computer Vision - Extracting Data from Images\n\n```elixir\nMix.install(\n  [\n    {:instructor_lite, path: Path.expand(\"../../\", __DIR__)},\n    {:req, \"~> 0.5\"},\n    {:kino, \"~> 0.12.3\"}\n  ]\n)\n```","ref":"vision.html"},{"type":"extras","title":"Motivation - Computer Vision - Extracting Data from Images","doc":"In recent months, the latest AI researcher labs have turned LLMs into multimodal models. What this means is that no longer do they just interpret text, but they can also interpret images. One example of this provided by Anthropic is the Clause 3.5 Sonnet model. With no extra work, you can now provide images into your prompts with instructor and still do the normal structured extractions that you're used to.\n\nIn the following example, we will extract product details from a screenshot of a Shopify store.","ref":"vision.html#motivation"},{"type":"extras","title":"Setup - Computer Vision - Extracting Data from Images","doc":"In order to run code in this notebook, you need to add your Anthropic API key as an `OPENAI_KEY` Livebook secret. It will then be accessible through an environmental variable.\n\n```elixir\nsecret_key = System.fetch_env!(\"LB_ANTHROPIC_KEY\")\n:ok\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n:ok\n```","ref":"vision.html#setup"},{"type":"extras","title":"Example - Computer Vision - Extracting Data from Images","doc":"![](files/shopify-screenshot.png)\n\n```elixir\nimage = Kino.FS.file_path(\"shopify-screenshot.png\") |> File.read!()\nbase64_image = Base.encode64(image)\n\ndefmodule Product do\n  use Ecto.Schema\n\n  @primary_key false\n  embedded_schema do\n    field(:name, :string)\n    field(:price, :decimal)\n    field(:currency, Ecto.Enum, values: [:usd, :gbp, :eur, :cny])\n    field(:color, :string)\n  end\nend\n\n{:ok, result} =\n  InstructorLite.instruct(%{\n      messages: [\n        %{\n          role: \"user\",\n          content: [\n            %{type: \"text\", text: \"What is the product details of the following image?\"},\n            %{type: \"image\", source: %{data: base64_image, type: \"base64\", media_type: \"image/png\"}}\n          ]\n        }\n      ]\n    },\n    adapter: InstructorLite.Adapters.Anthropic,\n    response_model: Product,\n    adapter_context: [api_key: secret_key]\n  )\n\nresult\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Product{\n  name: \"Thomas Wooden Railway Thomas The Tank Engine\",\n  price: Decimal.new(\"33.0\"),\n  currency: :usd,\n  color: \"blue\"\n}\n```\n\n<!-- livebook:{\"offset\":2166,\"stamp\":{\"token\":\"XCP.VFIWnxLfdgy2x6C34CJrkmOCjlrQDs8J5R9CYnv-v6v2BLmiGtowOqI5wULoyB5eEn27o1QWQg2KsPu3_FGBRzNttsvwhcpnSWDN2Ie_ZShEOsRVdxKLtHCWzmo51J9Og2Rmee5NYA\",\"version\":2}} -->","ref":"vision.html#example"},{"type":"extras","title":"Converting Text into Explorer DataFrames","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Converting Text into Explorer DataFrames\n\n```elixir\nMix.install(\n  [\n    {:instructor_lite, path: Path.expand(\"../../\", __DIR__)},\n    {:req, \"~> 0.5\"},\n    {:explorer, \"~> 0.7.2\"},\n    {:kino, \"~> 0.12.0\"},\n    {:kino_explorer, \"~> 0.1.13\"}\n  ]\n)\n```","ref":"text-to-dataframes.html"},{"type":"extras","title":"Motivation - Converting Text into Explorer DataFrames","doc":"Sometimes we come across text that can be naturally extracted into structured data. Although we don't necessarily know how many tables, and what the schemas are represented in the data. Using Instructor we can let the LLM determine the tables and schemas for us, pulling the results into a dataframe which we can then use for further analysis.","ref":"text-to-dataframes.html#motivation"},{"type":"extras","title":"Setup - Converting Text into Explorer DataFrames","doc":"In order to run code in this notebook, you need to add your OpenAI API key as an `OPENAI_KEY` Livebook secret. It will then be accessible through an environmental variable.\n\n```elixir\nsecret_key = System.fetch_env!(\"LB_OPENAI_KEY\")\n:ok\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n:ok\n```","ref":"text-to-dataframes.html#setup"},{"type":"extras","title":"The Schema - Converting Text into Explorer DataFrames","doc":"Notice from this example we want resulting data structure to have data frames. That's not a default Ecto type so we're gonna have to do things a little different here.\n\nLuckily, Ecto allows us to define our own custom types in which we can override how the values are casted.\n\nIn our case, we're going to create a type for a table of data, which will be represented as a CSV string with respect to the LLM. Our custom type will then parse the string when it is casted by Ecto and parse the CSV into an `Explorer.DataFrame`.\n\n```elixir\ndefmodule Ecto.CSVDataFrame do\n  use Ecto.Type\n\n  def type, do: :string\n\n  def cast(csv_str) when is_binary(csv_str) do\n    df = Explorer.DataFrame.load_csv!(csv_str)\n    {:ok, df}\n  end\n\n  def cast(%Explorer.DataFrame{} = df), do: {:ok, df}\n  def cast(_), do: :error\n\n  def dump(x), do: {:ok, x}\n  def load(x), do: {:ok, x}\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Ecto.CSVDataFrame, <<70, 79, 82, 49, 0, 0, 11, ...>>, {:load, 1}}\n```\n\nNow that we have our data type, we can use it in our embedded Ecto schema, just as we would have in any other example. The built-in `InstructorLite.JSONSchema` module won't help us here, as it doesn't know how to convert a custom Ecto type, so we'll craft our own schema in the `json_schema` callback. If everything works correctly, we should get some data frames out.\n\n```elixir\ndefmodule Database do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @notes \"\"\"\n  The extracted database will contain one or more tables with data as csv formatted with ',' delimiters\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    embeds_many :tables, DataFrame, primary_key: false do\n      field(:name, :string)\n      field(:data, Ecto.CSVDataFrame)\n    end\n  end\n\n  @impl InstructorLite.Instruction\n  def json_schema() do\n    %{\n      type: \"object\",\n      description: \"\",\n      title: \"Database\",\n      required: [:tables],\n      \"$defs\": %{\n        \"DataFrame\" => %{\n          type: \"object\",\n          description: \"\",\n          title: \"DataFrame\",\n          required: [:data, :name],\n          properties: %{data: %{type: \"string\"}, name: %{type: \"string\"}},\n          additionalProperties: false\n        }\n      },\n      properties: %{tables: %{type: \"array\", items: %{\"$ref\": \"#/$defs/DataFrame\"}}},\n      additionalProperties: false\n    }\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Database, <<70, 79, 82, 49, 0, 0, 18, ...>>, {:json_schema, 0}}\n```\n\nPerfect. And now let's test it out on some example text.\n\n```elixir\nextract_df = fn text ->\n  InstructorLite.instruct(%{\n      model: \"gpt-4o-mini\",\n      messages: [\n        %{\n          role: \"system\",\n          content: \"\"\"\n          Map this data into one or more dataframes and correctly define the columns and data\n          \"\"\"\n        },\n        %{\n          role: \"user\",\n          content: \"#{text}\"\n        }\n      ]\n    },\n    response_model: Database,\n    adapter_context: [api_key: secret_key]\n  )\nend\n\n{:ok, db} =\n  extract_df.(\"\"\"\n    My name is John and I am 25 years old. I live in\n    New York and I like to play basketball. His name is\n    Mike and he is 30 years old. He lives in San Francisco\n    and he likes to play baseball. Sarah is 20 years old\n    and she lives in Los Angeles. She likes to play tennis.\n    Her name is Mary and she is 35 years old.\n    She lives in Chicago.\n    \n    On one team 'Tigers' the captain is John and there are 12 players.\n    On the other team 'Lions' the captain is Mike and there are 10 players.\n  \"\"\")\n\nKino.Layout.tabs(\n  for table <- db.tables do\n    {table.name, table.data}\n  end\n)\n```\n\n<!-- livebook:{\"offset\":4583,\"stamp\":{\"token\":\"XCP.xSeGz27KbDBtRKp3G_PZKpeEJWeiZtuV7WEgJv3LyA1qiE1lqhNSlcksitBtnvXCxZ8625YEHbHd-sbiAXRsh6gsXIIsnVhD4JjQfFD_Wh8o-1AAvQ\",\"version\":2}} -->","ref":"text-to-dataframes.html#the-schema"},{"type":"extras","title":"Using OpenAI Batch API","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Using OpenAI Batch API\n\n```elixir\nMix.install(\n  [\n    {:instructor_lite, path: Path.expand(\"../../\", __DIR__)},\n    {:req, \"~> 0.5\"},\n    {:multipart, \"0.4.0\"}\n  ]\n)\n```","ref":"batch-api.html"},{"type":"extras","title":"Motivation - Using OpenAI Batch API","doc":"OpenAI offers asynchronous [Batch API](https://platform.openai.com/docs/guides/batch) at a 50% discount. This can make a huge difference on the bill if we have a lot of workload that is done in the background and isn't time-sensitive.","ref":"batch-api.html#motivation"},{"type":"extras","title":"Setup - Using OpenAI Batch API","doc":"In order to run code in this notebook, you need to add your OpenAI API key as an `OPENAI_KEY` Livebook secret. It will then be accessible through an environment variable.\n\n```elixir\nsecret_key = System.fetch_env!(\"LB_OPENAI_KEY\")\n:ok\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n:ok\n```","ref":"batch-api.html#setup"},{"type":"extras","title":"Classifying All the Emails - Using OpenAI Batch API","doc":"We'll use our classic email classification example, but this time for a bunch of emails at once.\n\n```elixir\ndefmodule SpamPrediction do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @notes \"\"\"","ref":"batch-api.html#classifying-all-the-emails"},{"type":"extras","title":"Field Descriptions: - Using OpenAI Batch API","doc":"- class: Whether or not the email is spam.\n  - reason: A short, less than 10 word rationalization for the classification.\n  - score: A confidence score between 0.0 and 1.0 for the classification.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl true\n  def validate_changeset(changeset, _opts) do\n    changeset\n    |> Ecto.Changeset.validate_number(:score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, SpamPrediction, <<70, 79, 82, 49, 0, 0, 19, ...>>, {:validate_changeset, 2}}\n```\n\n```elixir\nopts = [\n  response_model: SpamPrediction,\n  adapter: InstructorLite.Adapters.OpenAI\n]\n\nprompts =\n  [\n    \"Hello I am a Nigerian prince and I would like to send you money\",\n    \"Hello I know we never met but I got access to your webcam and have some spicy footage\",\n    \"Elixir Radar Newsletter. Issue #420\"\n  ]\n  |> Enum.map(fn text ->\n    InstructorLite.prepare_prompt(%{\n        model: \"gpt-4o-mini\",\n        messages: [%{\n          role: \"user\",\n          content: \"Classify first paragraph of an email:\\n#{text}\"\n        }]\n      },\n      opts\n    )\n  end)\n  |> Enum.with_index()\n\njsonl =\n  prompts\n  |> Enum.map(fn {prompt, index} ->\n    Jason.encode!(%{custom_id: \"#{index}\", method: \"POST\", url: \"/v1/chat/completions\", body: prompt})\n  end)\n  |> Enum.join(\"\\n\")\n\nmultipart =\n  Multipart.new()\n  |> Multipart.add_part(Multipart.Part.text_field(\"batch\", \"purpose\"))\n  |> Multipart.add_part(\n    Multipart.Part.file_content_field(\"emails.jsonl\", jsonl, :file, filename: \"emails.jsonl\")\n  )\n\nauth_headers = [{\"Authorization\", \"Bearer #{secret_key}\"}]\nmp_headers = [{\"Content-Type\", Multipart.content_type(multipart, \"multipart/form-data\")}]\n\n{:ok, %{body: %{\"id\" => file_id}}} =\n  Req.post(\"https://api.openai.com/v1/files\",\n    headers: auth_headers ++ mp_headers,\n    body: Multipart.body_binary(multipart)\n  )\n\n{:ok, %{body: %{\"id\" => batch_id}}} =\n  Req.post(\"https://api.openai.com/v1/batches\",\n    json: %{input_file_id: file_id, endpoint: \"/v1/chat/completions\", completion_window: \"24h\"},\n    headers: auth_headers\n  )\n\n:ok\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n:ok\n```\n\nNow we need to wait a little and fetch the batch result.\n\n```elixir\n{:ok, %{body: %{\"output_file_id\" => output_file_id}}} =\n  Req.get(\"https://api.openai.com/v1/batches/#{batch_id}\", headers: auth_headers)\n\n{:ok, %{body: body}} =\n  Req.get(\"https://api.openai.com/v1/files/#{output_file_id}/content\", headers: auth_headers)\n\nresult =\n  body\n  |> String.split(\"\\n\", trim: true)\n  |> Enum.map(&Jason.decode!/1)\n  |> Enum.zip(prompts)\n  |> Enum.map(fn {%{\"response\" => %{\"body\" => response}}, prompt} ->\n    InstructorLite.consume_response(response, prompt, response_model: SpamPrediction)\n  end)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[\n  ok: %SpamPrediction{class: :spam, reason: \"Common spam trope of money scams\", score: 0.95},\n  ok: %SpamPrediction{class: :spam, reason: \"Threatening language and webcam mention\", score: 0.95},\n  ok: %SpamPrediction{\n    class: :not_spam,\n    reason: \"Newsletter title suggests legitimate content\",\n    score: 0.9\n  }\n]\n```\n\n<!-- livebook:{\"offset\":4290,\"stamp\":{\"token\":\"XCP.0u8LUVJewKQ1FRMqky47zdPB_1pX8-2jBURg4PqSfvs4FfION2QCgXQqSZgtky3XUOYUwm7Df4SUIudTHnVIr5h-HaK0mUMMj63NriP32u6GVhtTIA\",\"version\":2}} -->","ref":"batch-api.html#field-descriptions"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}